"""Apply patches to a parsed json document.

Add { op: "add", path: "/a/b/c", value: v }
  Add value v to /a/b at index c. If /a/b is a list, c must be an integer, and
  v will be inserted at c. To append a value to list, use a trailing / in the
  path.

Remove { op: "remove", path: "/a/b/c" }
  Remove the key c from dictionary /a/b, or the item at index c from list /a/b.

Replace { op: "replace", path: "/a/b/c", value: v, src: k }
  Replace the item at c on /a/b with v. If check is specified, the patch will
  fail if the existing value at c is not k.

Merge { op: "merge", path: "/a/b/c", value: { k0: v0, k1: v1, ..., kn: vn } }
  This will create or replace all k0 - kn on c. It can be used to group
  multiple adds and replaces, or can act as an upsert.

Copy { op: "copy", path: "/b/c/d", src: "/a/b/c" }
  Copy the value at /a/b/c to d on /b/c.

Move { op: "move", path: "/b/c/d", src: "/a/b/c" }
  Equivalent to a copy from 'src' to 'path' followed by a remove at 'path'.

Test { op: "test", path: "/a/b/c" [, value: v] }
  Test that the index c on /a/b is specified. If value is given, check
  additionally that the value at c on /a/b equals v

Set Remove { op: "setremove", path: "/a/b", value: v }
  Remove v from the list at /a/b, if v is in /a/b.

Set Add { op: "setadd", path: "/a/b", value: v }
  Add v to the list at /a/b if it doesn't already exist in the list.

All paths must conform to RFC 6901.

Inspiration from http://jsonpatch.com with additional features to handle
certain types of concurrent editing.
"""
from collections import namedtuple
from bidon.util import try_parse_int


_NO_VAL = object()
_PATH_SEP = "/"
_Patch = namedtuple("Patch", ["op", "path", "value", "src"])


def Patch(op, path, value=_NO_VAL, src=_NO_VAL): # pylint: disable=invalid-name
  """Build a Patch tuple, with some defaults for optional values."""
  return _Patch(op=op, path=path, value=value, src=src)


def add(parent, idx, value):
  """Add a value to a dict."""
  if isinstance(parent, dict):
    if idx in parent:
      raise Exception("Item already exists")
    parent[idx] = value
  elif isinstance(parent, list):
    if idx == "" or idx == "~":
      parent.append(value)
    else:
      parent.insert(int(idx), value)
  else:
    raise Exception("Invalid path for operation")


def remove(parent, idx):
  """Remove a value from a dict."""
  if isinstance(parent, dict):
    del parent[idx]
  elif isinstance(parent, list):
    del parent[int(idx)]
  else:
    raise Exception("Invalid path for operation")


def replace(parent, idx, value, check_value=_NO_VAL):
  """Replace a value in a dict."""
  if isinstance(parent, dict):
    if idx not in parent:
      raise Exception("Item does not exist")
  elif isinstance(parent, list):
    idx = int(idx)
    if idx < 0 or idx >= len(parent):
      raise Exception("List index out of range")
  if check_value is not _NO_VAL:
    if parent[idx] != check_value:
      raise Exception("Check value did not pass")
  parent[idx] = value


def merge(parent, idx, value):
  """Merge a value."""
  target = get_child(parent, idx)
  for key, val in value.items():
    target[key] = val


def copy(src_parent, src_idx, dest_parent, dest_idx):
  """Copy an item."""
  if isinstance(dest_parent, list):
    dest_idx = int(dest_idx)
  dest_parent[dest_idx] = get_child(src_parent, src_idx)


def move(src_parent, src_idx, dest_parent, dest_idx):
  """Move an item."""
  copy(src_parent, src_idx, dest_parent, dest_idx)
  remove(src_parent, src_idx)


def test(parent, idx, value=_NO_VAL):
  """Check to see if an item exists."""
  try:
    val = get_child(parent, idx)
  except Exception:
    return False

  if value is _NO_VAL:
    return True
  return val == value


def set_remove(parent, idx, value):
  """Remove an item from a list."""
  lst = get_child(parent, idx)
  if value in lst:
    lst.remove(value)


def set_add(parent, idx, value):
  """Add an item to a list if it doesn't exist."""
  lst = get_child(parent, idx)
  if value not in lst:
    lst.append(value)


def get_children(parent, idx):
  """Gets the child at parent[idx], or all the children if idx == "*"."""
  if isinstance(parent, dict):
    if idx in parent:
      yield parent[idx]
    else:
      raise Exception("Invalid path at {0}".format(idx))

  elif isinstance(parent, list):
    if idx == "*":
      yield from parent
    else:
      is_int, i = try_parse_int(idx)

      if is_int and i >= 0 and i < len(parent):
        yield parent[i]
      else:
        raise Exception("Invalid list index: {0}".format(i))

  else:
    raise Exception("Type {0} does not have children".format(type(parent).__name__))


def get_child(parent, idx):
  """Get the first child according to idx."""
  return next(get_children(parent, idx))


def parse_path(path):
  """Parse a rfc 6901 path."""
  if not path:
    raise ValueError("Invalid path")

  if isinstance(path, str):
    if path == "/":
      raise ValueError("Invalid path")
    if path[0] != "/":
      raise ValueError("Invalid path")
    return path.split(_PATH_SEP)[1:]
  elif isinstance(path, (tuple, list)):
    return path
  else:
    raise ValueError("A path must be a string, tuple or list")


def resolve_path(root, path):
  """Resolve a rfc 6901 path, returning the parent and the last path part."""
  path = parse_path(path)

  parent = root

  for part in path[:-1]:
    parent = get_child(parent, rfc_6901_replace(part))

  return (parent, rfc_6901_replace(path[-1]))


def find(root, path):
  """Get the (first) child at path from root."""
  return get_child(*resolve_path(root, path))


def find_all(root, path):
  """Get all children that satisfy the path."""
  path = parse_path(path)

  if len(path) == 1:
    yield from get_children(root, path[0])
  else:
    for child in get_children(root, path[0]):
      yield from find_all(child, path[1:])


def rfc_6901_replace(path):
  """Implements rfc 6901 escape code replacement."""
  return path.replace("~1", "/").replace("~0", "~")


def apply_patch(document, patch):
  """Apply a Patch object to a document."""
  # pylint: disable=too-many-return-statements
  op = patch.op
  parent, idx = resolve_path(document, patch.path)
  if op == "add":
    return add(parent, idx, patch.value)
  elif op == "remove":
    return remove(parent, idx)
  elif op == "replace":
    return replace(parent, idx, patch.value, patch.src)
  elif op == "merge":
    return merge(parent, idx, patch.value)
  elif op == "copy":
    sparent, sidx = resolve_path(document, patch.src)
    return copy(sparent, sidx, parent, idx)
  elif op == "move":
    sparent, sidx = resolve_path(document, patch.src)
    return move(sparent, sidx, parent, idx)
  elif op == "test":
    return test(parent, idx, patch.value)
  elif op == "setremove":
    return set_remove(parent, idx, patch.value)
  elif op == "setadd":
    return set_add(parent, idx, patch.value)
  else:
    raise Exception("Invalid operator")


def apply_patches(document, patches):
  """Serially apply all patches to a document."""
  for i, patch in enumerate(patches):
    try:
      result = apply_patch(document, patch)
      if patch.op == "test" and result is False:
        raise Exception("Test patch {0} failed. Cancelling entire set.".format(i + 1))
    except Exception as ex:
      raise Exception("An error occurred with patch {0}: {1}".format(i + 1, ex)) from ex



"""Contains the DataTable class and some helper functions."""
import re
from decimal import Decimal
from numbers import Number

from bidon.util.convert import to_compressed_string


_REPEATING_NUMBER_TRIM_RE = re.compile(r"((\d+?)\2{2,})")


def reduce_number(num):
  """Reduces the string representation of a number.

  If the number is of the format n.00..., returns n.
  If the decimal portion of the number has a repeating decimal, followed by up to two trailing
  numbers, such as:

  0.3333333

  or

  0.343434346

  It will return just one instance of the repeating decimals:

  0.3

  or

  0.34
  """
  parts = str(num).split(".")
  if len(parts) == 1 or parts[1] == "0":
    return int(parts[0])
  else:
    match = _REPEATING_NUMBER_TRIM_RE.search(parts[1])
    if match:
      from_index, _ = match.span()
      if from_index == 0 and match.group(2) == "0":
        return int(parts[0])
      else:
        return Decimal(parts[0] + "." + parts[1][:from_index] + match.group(2))
    else:
      return num


class DataTable(object):
  """This class contains methods for working with a table a cells. A cell is any object that holds
  values.

  Methods for interacting with the cells are passed in on the constructor."""
  def __init__(self, rows, get_cell_value=None, set_cell_value=None, is_cell_empty=None,
               clean_value=None):
    """Initialize the DataTable instance."""
    self.rows = rows
    self.nrows = len(rows)
    self.ncols = max([len(row) for row in rows]) if self.nrows != 0 else 0
    self._get_cell_value = get_cell_value
    self._set_cell_value = set_cell_value
    self._is_cell_empty = is_cell_empty
    self._clean_value = clean_value

  # TODO: coords is (col, row) in spreadsheet but (row, col) here. That should be changed...?"""
  def __getitem__(self, index):
    """Get the item at the given index.

    Index is a tuple of (row, col)
    """
    row, col = index
    return self.rows[row][col]

  def get_cell_value(self, cell):
    """Gets the value from the given cell.
    """
    if self._get_cell_value:
      return self._get_cell_value(cell)
    else:
      return cell

  def set_cell_value(self, row, index, value):
    """Sets the value for the cell at row[index]."""
    if self._set_cell_value:
      self._set_cell_value(row, index, value)
    else:
      row[index] = value

  def is_cell_empty(self, cell):
    """Checks if the cell is empty."""
    if cell is None:
      return True
    elif self._is_cell_empty:
      return self._is_cell_empty(cell)
    else:
      return cell is None

  def is_row_empty(self, row):
    """Returns True if every cell in the row is empty."""
    for cell in row:
      if not self.is_cell_empty(cell):
        return False
    return True

  def serialize(self, serialize_cell=None):
    """Returns a list of all rows, with serialize_cell or self.get_cell_value called on the cells of
    each.
    """
    if serialize_cell is None:
      serialize_cell = self.get_cell_value
    return [[serialize_cell(cell) for cell in row] for row in self.rows]

  def headers(self, serialize_cell=None):
    """Gets the first row of the data table, with serialize_cell or self.get_cell_value is called on
    each cell."""
    if serialize_cell is None:
      serialize_cell = self.get_cell_value
    return [serialize_cell(cell) for cell in self.rows[0]]

  def rows_to_dicts(self, serialize_cell=None):
    """Generates a sequence of dictionaries of {header[i] => row[i]} for each row."""
    if serialize_cell is None:
      serialize_cell = self.get_cell_value
    # keys = [serialize_cell(cell) for cell in self.rows[0]]
    keys = self.headers(serialize_cell)
    for row in self.rows[1:]:
      yield dict(zip(keys, [serialize_cell(cell) for cell in row]))

  def cleanup(self):
    """Cleans up the data table:

    It first calls clean_value passed to the constructor on each cell. Then it trims out empty
    rows and columns.
    """
    self.clean_values()
    self.trim_empty_rows()
    self.trim_empty_columns()
    return self

  def trim_empty_rows(self):
    """Remove all trailing empty rows."""
    if self.nrows != 0:
      row_index = 0
      for row_index, row in enumerate(reversed(self.rows)):
        if not self.is_row_empty(row):
          break
      self.nrows = len(self.rows) - row_index
      self.rows = self.rows[:self.nrows]
    return self

  def trim_empty_columns(self):
    """Removes all trailing empty columns."""
    if self.nrows != 0 and self.ncols != 0:
      last_col = -1
      for row in self.rows:
        for i in range(last_col + 1, len(row)):
          if not self.is_cell_empty(row[i]):
            last_col = i
      ncols = last_col + 1
      self.rows = [row[:ncols] for row in self.rows]
      self.ncols = ncols
    return self

  def clean_values(self):
    """Cleans the values in each cell. Calls either the user provided clean_value, or the
    class defined clean value.
    """
    for row in self.rows:
      for index, cell in enumerate(row):
        self.set_cell_value(row, index, self.clean_value(self.get_cell_value(cell)))
    return self

  def clean_value(self, value):
    """Cleans a value, using either the user provided clean_value, or cls.reduce_value."""
    if self._clean_value:
      return self._clean_value(value)
    else:
      return self.reduce_value(value)

  @classmethod
  def reduce_value(cls, value):
    """Cleans the value by either compressing it if it is a string, or reducing it if it is a
    number.
    """
    if isinstance(value, str):
      return to_compressed_string(value)
    elif isinstance(value, bool):
      return value
    elif isinstance(value, Number):
      return reduce_number(value)
    else:
      return value



import os
from pip.req import parse_requirements
from setuptools import setup


def get_requirements(filename):
  if not os.path.isabs(filename):
    filename = os.path.join(proj_root, filename)
  return [str(r.req) for r in parse_requirements(filename)]


proj_root = os.path.dirname(os.path.abspath(__file__))
version = "0.1.3"

setup(name="bidon",
      version=version,
      license="BSD",
      platforms="any",
      author="Trey Cucco",
      author_email="fcucco@gmail.com",
      packages=['bidon'],
      description="A simple, easy to use, and flexible data handling library",
      url="https://github.com/treycucco/bidon",
      download_url="https://github.com/treycucco/bidon/tarball/master",
      # install_requires=get_requirements("requirements.txt"),
      classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3 :: Only"
      ],
      package_data={"": ["*.txt"]})
